/*
 * This program is called by automount when an attempt is made to mount
 * a directory.  It is passed the directory name and will print out the
 * mount command for that directory.
 *
 * If the directory should not exist it does not produce any output at
 * all.
 *
 * We use the cwd as the location where this mount will occure.  I hope
 * this is correct.
 * 
 */

#include <stdio.h>
#include <stdarg.h>
#include <syslog.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <dirent.h>
#include <limits.h>
#include <string.h>
#include "../version.h"
#include "../smbpw.h"

#ifndef CONFIG_FILE
#define CONFIG_FILE "/etc/smbautomount.conf"
#endif

#define FILESYSTEM "cifs"

static void log(int err, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    openlog("smbautomount", LOG_CONS|LOG_PID, LOG_AUTH);
    vsyslog(err, format, args);
    va_end(args);
    closelog();
}


void usage(char *s)
{
	fprintf(stderr, "Usage: %s <username>\n", s);
}

/*
 * The config file (/etc/smbautomount.conf) contains lines like this
 * "/home:servername:options"
 *
 * If no match is found then localhost is returned
 */
int load_config(char *path, char **server, char **options)
{
	FILE *config;
	char buffer[4096];

	char *conf_path;
	char *conf_server;
	char *conf_option;

	int len;

	*server = "localhost";
	*options = "";

	config = fopen(CONFIG_FILE, "r");
	if(config)
	{
		while(fgets(buffer, sizeof(buffer), config))
		{
			// see if this line is a match...
			// strip the leading white space...
			conf_path = buffer;
			while(conf_path[0] == ' ' || conf_path[0] == '\t')
			{
				conf_path++;
			}

			// remove the last char if it is a \n
			len = strlen(buffer);
			if(len > 0 && buffer[len - 1] == '\n')
			{
				buffer[len - 1] = 0;
			}

			if(buffer[0] == '#')
			{
				// this is a comment
			}
			else
			{
				// look for the first ':'
				conf_server = strchr(conf_path, ':');
				if(!conf_server)
				{
					continue;
				}
				*conf_server = 0;
				conf_server++;

				// see if we have a match
				if(strcmp(path, conf_path) == 0)
				{
					// we have a match
					// find the options

					conf_option = strchr(conf_server, ':');
					if(conf_option)
					{
						*conf_option = 0;
						conf_option++;
					}

					// clone the results...
					*server = strdup(conf_server);
					if(conf_option)
					{
						*options = strdup(conf_option);
					}
					break;
					// and we are done
				}
			}
		}
		
		fclose(config);
	}
	return 0;
}

int main(int argc, char **argv)
{
	char buffer[4096];
	char home[4096];
	char ubuf[LINE_MAX];
	char ubuf2[LINE_MAX];
	char ubuf3[LINE_MAX];
	char *cwd;
	struct passwd passwd, *pwd;
	struct passwd passwd2, *pwd2;
	struct group group3, *grp3;
	char *host;
	//char *options = "";//"port=1139,";
	char *options;
	char *username;
	DIR *console_dir;
	struct dirent *console_dirent;
	char *dmask = "700";

	log(LOG_DEBUG, "smbautomount version "VERSION"\n");

	cwd = getcwd(buffer, sizeof(buffer));
	log(LOG_DEBUG, "cwd = %s\n", cwd);
	if(argc == 2)
	{
		username = argv[1];
		if (getpwnam_r(username, &passwd, ubuf, sizeof(ubuf), &pwd) != 0)
		{
			pwd = NULL;
		}
		if (pwd == NULL)
		{
			log(LOG_DEBUG, "user %s does not exist\n", username);
			return 0;
		}

		// check that this user's home directory is
		// $pwd/$user

		if((strlen(cwd) + strlen(username) + 1) < sizeof(home))
		{
			strcpy(home, cwd);
			strcat(home, "/");
			strcat(home, username);

			//if(strcmp(home, pwd->pw_dir) == 0)
			if(1)
			{
				load_config(cwd, &host, &options);
				// we still need some extra info for the template.  A config file....?
				//fprintf(stdout, "-fstype=smbfs,username=jn,port=1139,uid=jn,gid=jn,pwsock ://max/jn\n");
/*
				template = strdup("-fstype=smbfs,username=%n,%ouid=%u,gid=%g,pwsock ://%h/%n");

				i = 0;
				while(template[i])
				{
					if((template[i] == '%') && (template[i + 1] != 0))
					{
						i++;
						switch(template[i])
						{
							case 'n':
								printf("%s", argv[1]);
								break;

							case 'o':
								printf("%s", options);
								break;

							case 'u':
								printf("%d", pwd->pw_uid);
								break;

							case 'g':
								printf("%d", pwd->pw_gid);
								break;

							case 'h':
								printf("%s", host);
								break;

							default:
								log(LOG_ERR, "Invalid template parameter \%%c\n", template[i]);
								break;
						}
					}
					else
					{
						printf("%c", template[i]);
					}
					i++;
				}
				printf("\n");

*/

				if(!retrieve(username))
				{
					// The password is not available
					// Pick a new user to try...
					// what about the console owner?
					// /var/run/console
					console_dir = opendir("/var/run/console");
					if(console_dir)
					{
						while((console_dirent = readdir(console_dir)))
						{
							if(console_dirent->d_name[0] != '.')
							{

								// we should check that is is a user account name
								// see if there is a password
								if(retrieve(console_dirent->d_name))
								{
									//fprintf(stderr, "password for console owner %s is available\n", console_dirent->d_name);
									// lets give this user a try...
									//
									// but only if their group name = the desired username
									if (getpwnam_r(console_dirent->d_name, &passwd2, ubuf2, sizeof(ubuf2), &pwd2) == 0)
									{
										//fprintf(stderr, "gid is %d\n", pwd2->pw_gid);
										// the gid is passwd2->pw_gid
										if(getgrgid_r(pwd2->pw_gid, &group3, ubuf3, sizeof(ubuf3), &grp3) == 0)
										{
											// compare the strings
											//fprintf(stderr, "dir = %s user = %s group = %s\n", argv[1], console_dirent->d_name, grp3->gr_name);
											if(strcmp(argv[1], grp3->gr_name) == 0)
											{
												// it's a match
												username = strdup(console_dirent->d_name);
												dmask = "770";
												break;
											}
										}
									}
								}
							}
						};
						closedir(console_dir);
					}


				}

				// we should find out the correct dmask from somewhere.....
				// we want to allow group access.  These permissions will still be checked by the server
				// but we should try and make it look acceptable anyway.
				fprintf(stdout, "-fstype="FILESYSTEM",username=%s,%suid=%d,gid=%d,dmask=%s,fmask=600,pwsock ://%s/%s\n",
						username, options, pwd->pw_uid, pwd->pw_gid, dmask, host, argv[1]);
			}
			else
			{
				log(LOG_DEBUG, "paths do not match\n", argv[1]);
				log(LOG_DEBUG, "home   = %s\n", home);
				log(LOG_DEBUG, "pw_dir = %s\n", pwd->pw_dir);
			}

		}
		else
		{
			log(LOG_DEBUG, "path too long\n");
		}
	}
	else
	{
		usage(argv[0]);
	}
	return 0;
}
