/*
 * This is a passowrd cache for smb passwords
 *
 * PAM will register a password with the cache when a user logs in.
 * This password can then be retreived by smbmount when automount
 * requests their home directory be mounted.
 *
 * Communication is performed via a socket
 * /var/run/smbpw.sock
 *
 * Only root may connect to the socket.
 *
 * Simple commands are written to the socket and a simple
 * reply is returned.
 *
 * There are currently 3 commands supported:
 *
 * STORE:
 *   0x00, str(username), str(password)
 *
 *   result:
 *   0x00 = success
 *   0xnn = error
 *
 * RETRIEVE:
 *   0x01, str(username)
 *
 *   result
 *   0x01 = no data available
 *   0x00, str(password)
 *   0xnn where nn > 0x01 = error
 *
 * QUERY:
 *   0x03, str(username)
 *
 *   result
 *   0x01 = no data available
 *   0x00, str(NULL) = password available
 *   0xnn where nn > 0x01 = error
 * 
 * SHUTDOWN:
 *   0x02
 *
 *   result
 *   There is no result.  The server will terminate.
 *
 *
 * all str commands are 0x00 terminated string of chars.
 */

#include "version.h"
#include "smbpw.h"
#include <stdarg.h>
#include <syslog.h>

struct entry_t
{
	struct entry_t *next;
	char *key;
	char *value;
} *root = 0;

int cache_store_entry(char *key, char *value)
{
	// search from root to the end of the list...
	struct entry_t **entry = &root;
	struct entry_t *new;

	while(*entry)
	{
		if(strcmp((*entry)->key, key) == 0)
		{
			//printf("Already an entry for key %s\n", key);
			free((*entry)->value);
			(*entry)->value = strdup(value);
			return 1;
		}
		entry = &((*entry)->next);
	}
	// we are at the end of the list and the key does
	// not already exist
	new = malloc(sizeof(struct entry_t));
	new->next = 0;
	new->key = strdup(key);
	new->value = strdup(value);
	*entry = new;

	return 0;
}

char *cache_retrieve_entry(char *key)
{
	struct entry_t *entry = root;

	while(entry)
	{
		//printf("%s, %s\n", entry->key, entry->value);
		if(strcmp(key, entry->key) == 0)
		{
			return entry->value;
		}
		entry = entry->next;
	}
	return 0;
}

static void log(int err, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    openlog("smbpwman", LOG_CONS|LOG_PID, LOG_AUTH);
    vsyslog(err, format, args);
    va_end(args);
    closelog();
}

void print_version()
{
	fprintf(stderr, "smbpwman %s.  Written by John Newbigin <jn@it.swin.edu.au>\n", VERSION);
}

int main(int argc, char **argv)
{
	int ss;
	int cs;
	char command;
	int len;
	char *username;
	char *password;
	char result;
	int verbose = 0;
	FILE *lock;
	pid_t cpid;

	// verify that the socket path is secure...
	if(getuid() != 0)
	{
		print_version();
		fprintf(stderr, "Only root can run %s\n", argv[0]);
		return 1;
	}

	// create the lock file to make sure that this is the only process running...
	umask(0077);
	lock = fopen(PID_NAME, "w");
	if(lock == NULL)
	{
		print_version();
		fprintf(stderr, "Could not write lock file.  Is daemon already running?\n");
		return 1;
	}

	if(verify_socket_path(SOCKET_NAME) < 0)
	{
		print_version();
		fprintf(stderr, "Socket path %s is insecure\n", SOCKET_NAME);
		return 1;
	}

	ss = open_unix_socket(SOCKET_NAME);
	if(ss < 0)
	{
		print_version();
		fprintf(stderr, "Error creating socket\n");
		return 1;
	}

	// detach process...
	cpid=fork();

	if(cpid != 0)
	{
		exit(0);
	}
	setsid();
	chdir("/");

	fprintf(lock, "%d\n", getpid());
	//fclose(lock);
	close(0);
	close(1);
	close(2);


	log(LOG_INFO, "smbpwman daemon version %s ready\n", VERSION);
	while(1)
	{
		cs = accept(ss, 0, 0);
		if(cs == -1)
		{
			log(LOG_ERR, "smbpwman error %d on accept.  daemon terminating\n", errno);
			return 1;
		}
		
		len = read(cs, &command, sizeof(command));
		if(len > 0)
		{
			//printf("Got command %d\n", command);
		}
		else
		{
			close(cs);
			continue;
		}

		switch(command)
		{
			case 0x00:
				if(verbose)
				{
					printf("STORE\n");
				}
				username = readstring(cs);
				if(!username)
				{
					close(cs);
					continue;
				}
				password = readstring(cs);
				if(!password)
				{
					close(cs);
					continue;
				}
				log(LOG_DEBUG, "store request for %s\n", username);
				//printf("username = %s password = %s\n", username, password);
				cache_store_entry(username, password);
				result = 0x00;
				write(cs, &result, sizeof(result));
				break;

			case 0x01:
				if(verbose)
				{
					printf("RETRIEVE\n");
				}
				username = readstring(cs);
				if(!username)
				{
					close(cs);
					continue;
				}
				//printf("username = %s\n", username);
				log(LOG_DEBUG, "retrieve request for %s\n", username);
				password = cache_retrieve_entry(username);
				if(password)
				{
					result = 0x00;
					write(cs, &result, sizeof(result));
					write(cs, password, strlen(password) + 1);
				}
				else
				{
					result = 0x01;
					write(cs, &result, sizeof(result));
				}
				break;

			case 0x03:
				if(verbose)
				{
					printf("QUERY\n");
				}
				username = readstring(cs);
				if(!username)
				{
					close(cs);
					continue;
				}
				//printf("username = %s\n", username);
				log(LOG_DEBUG, "query request for %s\n", username);
				password = cache_retrieve_entry(username);
				if(password)
				{
					result = 0x00;
					write(cs, &result, sizeof(result));
					write(cs, "", 1);
				}
				else
				{
					result = 0x01;
					write(cs, &result, sizeof(result));
				}
				break;

			case 0x02:
				if(verbose)
				{
					printf("SHUTDOWN\n");
				}
				log(LOG_INFO, "shutdown request.  daemon terminating\n");
				return 0;
				break;

			default:
				break;
		}

		close(cs);
	}
	return 0;
}
